#include <GL/glew.h>
#include <gmath/gmath.h>

#include "mesh-gl.h"

MeshGL::MeshGL()
{
	vao = 0;

	vbo_vertices = 0;
	vbo_normals = 0;
	vbo_tex_coords = 0;
	ibo = 0;

	num_vertices = 0;
	num_indices = 0;
}

MeshGL::MeshGL(const MeshGL &mesh)
{
	indices = mesh.indices;
	vertices = mesh.vertices;
	normals = mesh.normals;
	tex_coords = mesh.tex_coords;

	vbo_vertices = 0;
	vbo_normals = 0;
	vbo_tex_coords = 0;
	ibo = 0;
	vao = 0;

	/*
	 * if we set these to the actual
	 * vertices.size() and indices.size()
	 * update_vbo will have no effect
	 * */

	num_vertices = 0;
	num_indices = 0;
}

MeshGL &MeshGL::operator=(const MeshGL &mesh)
{
	if(this == &mesh)
		return *this;

	/* to avoid OpenGL leaks */
	destroy_vbo();

	/* what the copy constructor does */
	indices = mesh.indices;
	vertices = mesh.vertices;
	normals = mesh.normals;
	tex_coords = mesh.tex_coords;

	return *this;
}

MeshGL::~MeshGL()
{
	destroy_vbo();

	vertices.clear();
	normals.clear();
}

void MeshGL::draw() const
{
	glBindVertexArray(vao);

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ibo);
	glDrawElements(GL_TRIANGLES, num_indices, GL_UNSIGNED_SHORT, 0);

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
	glBindVertexArray(0);
}

void MeshGL::update_vertex_data()
{
	update_vbo();
}

void MeshGL::update_vbo()
{
	if(vertices.empty()) {
		printf("empty vertices\n");
		return;
	}

	/* vao */
	if(!vao)
		glGenVertexArrays(1, &vao);
	glBindVertexArray(vao);

	/* vertices */

	if(!vbo_vertices)
		glGenBuffers(1, &vbo_vertices);
	glBindBuffer(GL_ARRAY_BUFFER, vbo_vertices);
	if(num_vertices != (int)vertices.size())
		glBufferData(GL_ARRAY_BUFFER, vertices.size() * sizeof(Vec3),
		             &vertices[0], GL_STATIC_DRAW);
	else
		glBufferSubData(GL_ARRAY_BUFFER, 0, vertices.size() * sizeof(Vec3),
		                &vertices[0]);
	glVertexAttribPointer(MESH_VERTEX, 3, GL_FLOAT, GL_FALSE, sizeof(Vec3), 0);

	/* normals */

	if(!vbo_normals)
		glGenBuffers(1, &vbo_normals);
	glBindBuffer(GL_ARRAY_BUFFER, vbo_normals);
	if(num_vertices != (int)normals.size())
		glBufferData(GL_ARRAY_BUFFER, normals.size() * sizeof(Vec3),
		             &normals[0], GL_STATIC_DRAW);
	else
		glBufferSubData(GL_ARRAY_BUFFER, 0, normals.size() * sizeof(Vec3),
		                &normals[0]);
	glVertexAttribPointer(MESH_NORMAL, 3, GL_FLOAT, GL_FALSE, sizeof(Vec3), 0);

	/* texture coordinates */

	if(!vbo_tex_coords)
		glGenBuffers(1, &vbo_tex_coords);
	glBindBuffer(GL_ARRAY_BUFFER, vbo_tex_coords);
	if(num_vertices != (int)tex_coords.size())
		glBufferData(GL_ARRAY_BUFFER, tex_coords.size() * sizeof(Vec2), &tex_coords[0], GL_STATIC_DRAW);
	else
		glBufferSubData(GL_ARRAY_BUFFER, 0, tex_coords.size() * sizeof(Vec2), &tex_coords[0]);
	glVertexAttribPointer(MESH_TEXTURE, 2, GL_FLOAT, GL_FALSE, sizeof(Vec2), 0);

	num_vertices = vertices.size();

	/* indices */

	if(!ibo)
		glGenBuffers(1, &ibo);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ibo);
	if(num_indices != (int)indices.size())
		glBufferData(GL_ELEMENT_ARRAY_BUFFER, indices.size() * 2,
		             &indices[0], GL_STATIC_DRAW);
	else
		glBufferSubData(GL_ELEMENT_ARRAY_BUFFER, 0, indices.size() * 2,
		                &indices[0]);

	num_indices = indices.size();

	glEnableVertexAttribArray(MESH_VERTEX);
	glEnableVertexAttribArray(MESH_NORMAL);
	glEnableVertexAttribArray(MESH_TEXTURE);

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindVertexArray(0);
}

void MeshGL::destroy_vbo()
{
	if(vbo_vertices)
		glDeleteBuffers(1, &vbo_vertices);
	if(vbo_normals)
		glDeleteBuffers(1, &vbo_normals);
	if(vbo_tex_coords)
		glDeleteBuffers(1, &vbo_tex_coords);
	if(ibo)
		if(vao)
			glDeleteVertexArrays(1, &vao);
}