#include <GL/glew.h>
#include <gmath/gmath.h>

#include "global.h"
#include "gfxapi.h"

#include "camera.h"
#include "mesh.h"
#include "meshgen.h"
#include "object.h"
#include "renderer.h"
#include "scene.h"
#include "shader.h"
#include "shader_manager.h"
#include "texture.h"

/* skybox */

static Mesh *skymesh; /* it will be used by many renderers => no destroy */
static ShaderProgram *skyprog;
static int viewproj_loc;

Renderer::Renderer()
{
	scene = 0;
	camera = 0;
	sprog = 0;

	skytex = 0;
	dskytex = 0;
}

Renderer::~Renderer()
{
}

bool Renderer::create()
{
	if(!(sprog = sdr_man->create_shader_program("default.v.glsl", "default.f.glsl"))) {
		return false;
	}

	/* getting material uniform locations: diffuse, specular, specular exponent (strength) */

	if((diff_loc = sprog->get_uniform_location("diffuse")) == -1) {
		fprintf(stderr, "Invalid uniform location: can't find \"diffuse\".\n");
		return false;
	}

	if((spec_loc = sprog->get_uniform_location("specular")) == -1) {
		fprintf(stderr, "Invalid uniform location: can't find \"specular\".\n");
		return false;
	}

	if((shin_loc = sprog->get_uniform_location("shininess")) == -1) {
		fprintf(stderr, "Invalid uniform location: can't find \"shininess\".\n");
		return false;
	}

	/* uniform locations for matrices */

	if((mmviewproj_loc = sprog->get_uniform_location("mmviewproj")) == -1) {
		fprintf(stderr, "Invalid uniform location: can't find \"mmviewproj\".\n");
		return false;
	}

	if((mview_loc = sprog->get_uniform_location("mview")) == -1) {
		fprintf(stderr, "Invalid uniform location: can't find \"mview\".\n");
		return false;
	}

	return true;
}

void Renderer::draw() const
{
	if(!camera || !scene)
		return;

	if(!sprog->link())
		return;

	sprog->use();

	for(size_t i=0; i<scene->objects.size(); i++) {
		draw_object(scene->objects[i]);
	}
}

void Renderer::draw_object(Object *object) const
{
	Material *m = object->material;

	/* setting uniforms for material */

	sprog->set_uniformf(diff_loc, m->diffuse.x, m->diffuse.y, m->diffuse.z, 1);
	sprog->set_uniformf(spec_loc, m->specular.x, m->specular.y, m->specular.z, 1);
	sprog->set_uniformf(shin_loc, m->shininess);

	/* texture */

	if(m->dtex)
		m->dtex->bind();

	/* setting uniforms for matrices */

	Mat4 mmviewproj = object->transform * camera->get_view_matrix() * mprojection;
	sprog->set_uniform_matrix(mmviewproj_loc, mmviewproj);

	Mat4 mview = object->transform * camera->get_view_matrix();
	sprog->set_uniform_matrix(mview_loc, mview);

	object->mesh->update_vertex_data();
	object->mesh->draw();
}

void Renderer::set_sky_tex(Texture *stex)
{
	if(!skymesh) {
		skymesh = gfx_create_mesh();
		gen_geosphere(skymesh, 1, 0);
	}
	if(!skyprog) {
		if(!(skyprog = sdr_man->create_shader_program("sky.v.glsl", "sky.f.glsl")))
			return;
		if((viewproj_loc = skyprog->get_uniform_location("viewproj") == -1))
			return;
	}
	skytex = stex;
}

void Renderer::set_diffuse_sky_tex(Texture *dstex)
{
	dskytex = dstex;
}

void Renderer::draw_skybox()
{
	if(!skymesh || !skytex)
		return;

	skytex->bind();
	skyprog->use();

	Mat4 mviewproj;
	mviewproj = camera->get_view_matrix().upper3x3() * mprojection;
	skyprog->set_uniform_matrix(viewproj_loc, mviewproj);

	skymesh->draw();
}